unit XmlIni;

interface

{
  Design By Howard Chen.
  EMail: howardchen@pie.com.tw
  MSN: powerhowardchen@hotmail.com

  XmlIni version 4.
}

uses
  Windows, Classes, XMLIntf, Dialogs;

type
  TXmlIni = class;
  TXmlIniSection = class;
  TXmlIniKey = class;

  TXmlIniEncodingType = (xietUnknow, xietUTF_8, xietUTF_16,
      xietBIG5, xietGB2312, xietS_JIS, xietEUC_KR, xietIS0_8859_1);

  TXmlIniDataType = (xidtText, xidtInteger, xidtFloat, xidtBool, xidtDateTime, xidtDate, xidtTime, xidtBinary);

  TXmlIniEventMemoryStream = class(TMemoryStream)
  private
    FOwner: TXmlIniKey;
  public
    constructor Create(AOwner: TXmlIniKey);
    function GetSize: Int64; override;
    procedure SetSize(NewSize: Longint); override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
  end;

  TXmlIniKey = class(TObject)
  private
    FOwner: TXmlIniSection;
    FName: WideString;
    FText: WideString;
    FStream: TXmlIniEventMemoryStream;
    FDataType: TXmlIniDataType;
    FRaiseException: Boolean;

    procedure BinaryToText;
    procedure TextToBinary;
    procedure SetName(const Value: WideString);
    function GetText: WideString;
    procedure SetText(const Value: WideString);
    function GetAsStream: TStream;
    procedure SetAsStream(const Value: TStream);
    function GetAsInteger: Integer;
    procedure SetAsInteger(const Value: Integer);
    function GetAsFloat: Double;
    procedure SetAsFloat(const Value: Double);
    function GetAsBool: Boolean;
    procedure SetAsBool(const Value: Boolean);
    function GetAsDateTime: TDateTime;
    procedure SetAsDateTime(const Value: TDateTime);
    function GetAsDate: TDateTime;
    procedure SetAsDate(const Value: TDateTime);
    function GetAsTime: TDateTime;
    procedure SetAsTime(const Value: TDateTime);
    procedure SetDataType(const Value: TXmlIniDataType);
  public
    constructor Create(AName: WideString = ''; AText: WideString = ''; AOwner: TXmlIniSection = nil);
    destructor Destroy; override;

    procedure Clear;

    procedure ValueLoadFromFile(FileName: WideString; Binary: Boolean = True); virtual;
    procedure ValueSaveToFile(FileName: WideString); virtual;

    property Name: WideString read FName write SetName;
    property Text: WideString read GetText write SetText;
    property AsStream: TStream read GetAsStream write SetAsStream;
    property AsInteger: Integer read GetAsInteger write SetAsInteger;
    property AsFloat: Double read GetAsFloat write SetAsFloat;
    property AsBool: Boolean read GetAsBool write SetAsBool;
    property AsDateTime: TDateTime read GetAsDateTime write SetAsDateTime;
    property AsDate: TDateTime read GetAsDate write SetAsDate;
    property AsTime: TDateTime read GetAsTime write SetAsTime;
    property DataType: TXmlIniDataType read FDataType write SetDataType;
    property RaiseException: Boolean read FRaiseException write FRaiseException;
  end;

  TXmlIniSection = class(TObject)
  private
    FName: WideString;
    FOwner: TXmlIni;
    FList: TList;
    FXmlEncoding: TXmlIniEncodingType;
    FCaseSensitive: Boolean;
    FCompressed: Boolean;
    FRaiseException: Boolean;
    
    function InternalFind(KeyName: WideString): TXmlIniKey;
    function InternalAdd(KeyName: WideString): TXmlIniKey;
    procedure SetName(const Value: WideString);
    function GetKeyByIndex(Index: Integer): TXmlIniKey;
    function GetKey(Index: WideString): TXmlIniKey;
    function GetItem(Index: WideString): WideString;
    procedure SetItem(Index: WideString; const Value: WideString);
    function GetCount: Integer;
    procedure SetRaiseException(const Value: Boolean);
  public
    constructor Create(AName: WideString = ''; AOwner: TXmlIni = nil);
    destructor Destroy; override;

    function Equals(Dest: TXmlIniSection): Boolean; overload;
    procedure CopyFrom(Source: TXmlIniSection); overload;
    procedure Append(Source: TXmlIniSection); overload;

    function Equals(Dest: TStrings): Boolean; overload;
    procedure CopyFrom(Source: TStrings); overload;
    procedure Append(Source: TStrings); overload;

    function Exists(KeyName: WideString): Boolean;

    procedure Clear;
    procedure Collation;

    procedure LoadFromStream(Stream: TStream; SectionName: WideString = ''); virtual;
    procedure SaveToStream(Stream: TStream; SectionName: WideString = ''); virtual;

    procedure LoadFromString(Value: WideString; SectionName: WideString = '');
    function SaveToString(SectionName: WideString = ''): WideString;

    procedure LoadFromBase64String(Value: string; SectionName: WideString = '');
    function SaveToBase64String(SectionName: WideString = ''): string;

    procedure LoadFromFile(FileName: WideString; SectionName: WideString = '');
    procedure SaveToFile(FileName: WideString; SectionName: WideString = '');

    procedure LoadFromRegistry(RootKey: HKEY; Path: WideString; SectionName: WideString = ''); virtual;
    procedure SaveToRegistry(RootKey: HKEY; Path: WideString; SectionName: WideString = '';
      UseOldDataType: Boolean = False); virtual;

    procedure LoadFromIni(FileName: WideString; SectionName: WideString = '');
    procedure SaveToIni(FileName: WideString; SectionName: WideString = '');

    property XmlEncoding: TXmlIniEncodingType read FXmlEncoding write FXmlEncoding;
    property CaseSensitive: Boolean read FCaseSensitive write FCaseSensitive;
    property Compressed: Boolean read FCompressed write FCompressed;

    property Name: WideString read FName write SetName;
    property KeyByIndex[Index: Integer]: TXmlIniKey read GetKeyByIndex;
    property Key[Index: WideString]: TXmlIniKey read GetKey;
    property Item[Index: WideString]: WideString read GetItem write SetItem; default;
    property Count: Integer read GetCount;
    property RaiseException: Boolean read FRaiseException write SetRaiseException;
  end;

  TXmlIni = class(TObject)
  private
    FList: TList;
    FXmlEncoding: TXmlIniEncodingType;
    FFileName: WideString;
    FAutoSave: Boolean;
    FCaseSensitive: Boolean;
    FCompressed: Boolean;
    FRaiseException: Boolean;

    procedure InternalXmlLoadFromStream(Stream: TStream);
    procedure InternalXmlSaveToStream(Stream: TStream);
    procedure InternalLoadFromFile(AFileName: WideString);
    procedure InternalSaveToFile(AFileName: WideString);
    function InternalFind(SectionName: WideString): TXmlIniSection;
    function InternalAdd(SectionName: WideString): TXmlIniSection;
    function GetCount: Integer;
    function GetSectionByIndex(Index: Integer): TXmlIniSection;
    function GetSection(Index: WideString): TXmlIniSection;
    procedure SetFileName(const Value: WideString);
    procedure SetAutoSave(const Value: Boolean);
    procedure SetCaseSensitive(const Value: Boolean);
    procedure SetCompressed(const Value: Boolean);
    procedure SetRaiseException(const Value: Boolean);
  public
    constructor Create(AFileName: WideString = '');
    destructor Destroy; override;

    function Equals(Dest: TXmlIni): Boolean;
    procedure CopyFrom(Source: TXmlIni);
    procedure Append(Source: TXmlIni);

    function Add(SectionName: WideString): TXmlIniSection;
    procedure Delete(SectionName: WideString);
    function Exists(SectionName: WideString; KeyName: WideString = ''): Boolean;

    procedure Clear;
    procedure Collation;

    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToStream(Stream: TStream); virtual;

    procedure LoadFromString(Value: WideString);
    function SaveToString: WideString;

    procedure LoadFromBase64String(Value: string);
    function SaveToBase64String: string;

    procedure LoadFromFile(AFileName: WideString = ''); virtual;
    procedure SaveToFile(AFileName: WideString = ''); virtual;

    procedure LoadFromRegistry(RootKey: HKEY; Path: WideString); virtual;
    procedure SaveToRegistry(RootKey: HKEY; Path: WideString); virtual;

    procedure LoadFromIni(AIniFileName: WideString); virtual;
    procedure SaveToIni(AIniFileName: WideString); virtual;

    procedure LoadFromShareMem(ShareMemID: string = ''); virtual;
    procedure SaveToShareMem(ShareMemID: string = ''); virtual;

    // TIniFile Functions Begin
    function SectionExists(const Section: WideString): Boolean;

    function ReadString(const Section, Ident, Default: WideString): WideString; virtual;
    procedure WriteString(const Section, Ident, Value: WideString); virtual;
    function ReadInteger(const Section, Ident: WideString; Default: Longint): Longint; virtual;
    procedure WriteInteger(const Section, Ident: WideString; Value: Longint); virtual;
    function ReadFloat(const Section, Ident: WideString; Default: Double): Double; virtual;
    procedure WriteFloat(const Section, Ident: WideString; Value: Double); virtual;
    function ReadBool(const Section, Ident: WideString; Default: Boolean): Boolean; virtual;
    procedure WriteBool(const Section, Ident: WideString; Value: Boolean); virtual;
    function ReadDateTime(const Section, Ident: WideString; Default: TDateTime): TDateTime; virtual;
    procedure WriteDateTime(const Section, Ident: WideString; Value: TDateTime); virtual;
    function ReadDate(const Section, Ident: WideString; Default: TDateTime): TDateTime; virtual;
    procedure WriteDate(const Section, Ident: WideString; Value: TDateTime); virtual;
    function ReadTime(const Section, Ident: WideString; Default: TDateTime): TDateTime; virtual;
    procedure WriteTime(const Section, Ident: WideString; Value: TDateTime); virtual;
    function ReadBinaryStream(const Section, Ident: WideString; Value: TStream): Integer; virtual;
    procedure WriteBinaryStream(const Section, Ident: WideString; Value: TStream); virtual;

    procedure ReadSection(const Section: WideString; Strings: TStrings); virtual;
    procedure ReadSections(Strings: TStrings); virtual;
    procedure ReadSectionValues(const Section: WideString; Strings: TStrings); virtual;
    procedure EraseSection(const Section: WideString); virtual;

    procedure DeleteKey(const Section, Ident: WideString); virtual;
    procedure UpdateFile; virtual;
    function ValueExists(const Section, Ident: WideString): Boolean;
    // TIniFile Functions End

    property FileName: WideString read FFileName write SetFileName;
    property AutoSave: Boolean read FAutoSave write SetAutoSave;

    property XmlEncoding: TXmlIniEncodingType read FXmlEncoding write FXmlEncoding;
    property CaseSensitive: Boolean read FCaseSensitive write SetCaseSensitive;
    property Compressed: Boolean read FCompressed write SetCompressed;

    property SectionByIndex[Index: Integer]: TXmlIniSection read GetSectionByIndex;
    property Section[Index: WideString]: TXmlIniSection read GetSection; default;
    property Count: Integer read GetCount;
    property RaiseException: Boolean read FRaiseException write SetRaiseException;
  end;

function ExStrToDateTime(const Value: WideString): TDateTime;
function ExStrToDate(const Value: WideString): TDateTime;
function ExStrToTime(const Value: WideString): TDateTime;
function ExDateTimeToStr(const Value: TDateTime): WideString;
function ExDateToStr(const Value: TDateTime): WideString;
function ExTimeToStr(const Value: TDateTime): WideString;

function Base64ToStr(Base64: WideString): WideString;
function StrToBase64(Str: WideString): WideString;

const
  XI_XML_EncodingStr: array [TXmlIniEncodingType] of WideString =
      ('Unknow', 'UTF-8', 'UTF-16', 'BIG5', 'GB2312', 'S-JIS', 'EUC-KR', 'IS0-8859-1');

implementation

uses
  xmldom, msxmldom, XMLDoc, ActiveX, Forms, zlib, SysUtils, DateUtils, StrUtils, Registry, IniFiles;

const
  XI_XML_StandAlone = 'yes';
  XI_XML_Version = '1.0';
  XI_XML_DefaultEncoding = xietUTF_16;

  XI_Root_NodeName = 'xmlini';
  XI_Section_NodeName = 'section';
  XI_Key_NodeName = 'key';
  XI_Name_Attribute = 'name';

  XI_Base64Header = '?BASE64?';

  XI_True_Text: array [0..14] of WideString =
      ('True', 'T', 'Yes', 'Y', 'Just', 'Right', 'Correct', 'OK', '1', 'On', 'Agree', 'O', '', 'n', 'T');
  XI_False_Text: array [0..17] of WideString =
      ('False', 'F', 'No', 'N', 'Unjust', 'Wrong', 'Incorrect', 'Cancel', '0', 'Off', 'Disagree', '_', '',
        '', '', 'n', 'T', '~');

  XI_LockMutexName = 'XMLINI_MUTEX_LOCK_';
  XI_ShareName = 'XMLINI_SHARED_MEM_';

const
  BufSize = 1024;

procedure CompressStream(SourceStream, DestStream: TStream);
var
  CS: TCompressionStream;
  Buf: PChar;
  DataSize, ReadCount: LongInt;
begin
  SourceStream.Position := 0;
  DestStream.Size := 0;

  DataSize := SourceStream.Size;
  DestStream.Write('ZL', 2);
  DestStream.Write(DataSize, SizeOf(LongInt));

  CS := TCompressionStream.Create(clMax, DestStream);
  try
    Buf := AllocMem(BufSize);
    try
      repeat
        ReadCount := SourceStream.Read(Buf^, BufSize);
        if ReadCount > 0 then
          CS.Write(Buf^, ReadCount);
      until ReadCount = 0;
    finally
      FreeMem(Buf);
    end;
  finally
    CS.Free;
  end;
end;

procedure DecompressStream(SourceStream, DestStream: TStream);
var
  DS: TDecompressionStream;
  Ident: array[0..1] of Char;
  Buf: PChar;
  DataSize, ReadCount: LongInt;
begin
  SourceStream.Position := 0;
  SourceStream.Read(Ident, 2);
  SourceStream.Read(DataSize, SizeOf(LongInt));
  if (Ident = 'ZL') and (DataSize > 0) then
  begin
    DestStream.Size := 0;
    DS := TDecompressionStream.Create(SourceStream);
    try
      Buf := AllocMem(BufSize);
      try
        while DataSize > 0 do
        begin
          ReadCount := DataSize;
          if ReadCount > BufSize then
            ReadCount := BufSize;
          DS.Read(Buf^, ReadCount);
          DestStream.Write(Buf^, ReadCount);
          Dec(DataSize, ReadCount);
        end;
      finally
        FreeMem(Buf);
      end;
    finally
      DS.Free;
    end;
  end;
end;

type
  PByte4 = ^TByte4;
  TByte4 = packed record
    b1: Byte;
    b2: Byte;
    b3: Byte;
    b4: Byte;
  end;

  PByte3 = ^TByte3;
  TByte3 = packed record
    b1: Byte;
    b2: Byte;
    b3: Byte;
  end;

const
  BUFFER_SIZE = $3000;
 { Caution: For MimeEncodeStream and all other kinds of multi-buffered
   Mime encodings (i.e. Files etc.), BufferSize must be set to a multiple of 3.
   Even though the implementation of the Mime decoding routines below
   does not require a particular buffer size, they work fastest with sizes of
   multiples of four. The chosen size is a multiple of 3 and of 4 as well.
   The following numbers are, in addition, also divisible by 1024:
   $2400, $3000, $3C00, $4800, $5400, $6000, $6C00. }

  MIME_ENCODE_TABLE: array[0..63] of Byte = (
    065, 066, 067, 068, 069, 070, 071, 072, // 00 - 07
    073, 074, 075, 076, 077, 078, 079, 080, // 08 - 15
    081, 082, 083, 084, 085, 086, 087, 088, // 16 - 23
    089, 090, 097, 098, 099, 100, 101, 102, // 24 - 31
    103, 104, 105, 106, 107, 108, 109, 110, // 32 - 39
    111, 112, 113, 114, 115, 116, 117, 118, // 40 - 47
    119, 120, 121, 122, 048, 049, 050, 051, // 48 - 55
    052, 053, 054, 055, 056, 057, 043, 047); // 56 - 63

  MIME_DECODE_TABLE: array[Byte] of Cardinal = (
    255, 255, 255, 255, 255, 255, 255, 255, //  00 -  07
    255, 255, 255, 255, 255, 255, 255, 255, //  08 -  15
    255, 255, 255, 255, 255, 255, 255, 255, //  16 -  23
    255, 255, 255, 255, 255, 255, 255, 255, //  24 -  31
    255, 255, 255, 255, 255, 255, 255, 255, //  32 -  39
    255, 255, 255, 062, 255, 255, 255, 063, //  40 -  47
    052, 053, 054, 055, 056, 057, 058, 059, //  48 -  55
    060, 061, 255, 255, 255, 255, 255, 255, //  56 -  63
    255, 000, 001, 002, 003, 004, 005, 006, //  64 -  71
    007, 008, 009, 010, 011, 012, 013, 014, //  72 -  79
    015, 016, 017, 018, 019, 020, 021, 022, //  80 -  87
    023, 024, 025, 255, 255, 255, 255, 255, //  88 -  95
    255, 026, 027, 028, 029, 030, 031, 032, //  96 - 103
    033, 034, 035, 036, 037, 038, 039, 040, // 104 - 111
    041, 042, 043, 044, 045, 046, 047, 048, // 112 - 119
    049, 050, 051, 255, 255, 255, 255, 255, // 120 - 127
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255);

function MimeEncodedSize(const i: Integer): Integer;
begin
  Result := (i + 2) div 3 * 4;
end;

procedure MimeEncode(const InputBuffer: Pointer; const InputByteCount: Integer; const OutputBuffer: Pointer);
var
  b: Cardinal;
  InMax3: Integer;
  pIn, PInLimit: ^Byte;
  POut: PByte4;
begin
{$IFDEF RangeChecking}
  if InputBuffer = nil then raise EMime.Create(SInputBufferNil);
  if OutputBuffer = nil then raise EMime.Create(SOutputBufferNil);
{$ENDIF}
  if InputByteCount <= 0 then Exit;

  pIn := InputBuffer;
  InMax3 := InputByteCount div 3 * 3;
  POut := OutputBuffer;

  Integer(PInLimit) := Integer(pIn) + InMax3;

  while pIn <> PInLimit do
  begin
    b := pIn^;
    b := b shl 8;
    Inc(pIn);
    b := b or pIn^;
    b := b shl 8;
    Inc(pIn);
    b := b or pIn^;
    Inc(pIn);

   // Write 4 bytes to OutputBuffer (in reverse order).
    POut.b4 := MIME_ENCODE_TABLE[b and $3F];
    b := b shr 6;
    POut.b3 := MIME_ENCODE_TABLE[b and $3F];
    b := b shr 6;
    POut.b2 := MIME_ENCODE_TABLE[b and $3F];
    b := b shr 6;
    POut.b1 := MIME_ENCODE_TABLE[b];

    Inc(POut);
  end;

  case InputByteCount - InMax3 of
    1:
      begin
        b := pIn^;

        b := b shl 4;

        POut.b2 := MIME_ENCODE_TABLE[b and $3F];
        b := b shr 6;
        POut.b1 := MIME_ENCODE_TABLE[b];

        POut.b3 := Ord('='); // Fill remaining 2 bytes.
        POut.b4 := Ord('=');
      end;
    2:
      begin
        b := pIn^;
        Inc(pIn);
        b := b shl 8;
        b := b or pIn^;

        b := b shl 2;

        POut.b3 := MIME_ENCODE_TABLE[b and $3F];
        b := b shr 6;
        POut.b2 := MIME_ENCODE_TABLE[b and $3F];
        b := b shr 6;
        POut.b1 := MIME_ENCODE_TABLE[b];

        POut.b4 := Ord('='); // Fill remaining byte.
      end;
  end;
end;

procedure MimeEncodeStream(const InputStream: TStream; const OutputStream: TStream);
var
  InputBuffer: array [0..BUFFER_SIZE - 1] of Byte;
  OutputBuffer: array [0..((BUFFER_SIZE + 2) div 3) * 4 - 1] of Byte;
  BytesRead: Integer;
begin
  BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  while BytesRead = SizeOf(InputBuffer) do
  begin
    MimeEncode(@InputBuffer, SizeOf(InputBuffer), @OutputBuffer);
    OutputStream.Write(OutputBuffer, SizeOf(OutputBuffer));
    BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  end;
  if BytesRead > 0 then
  begin
    MimeEncode(@InputBuffer, BytesRead, @OutputBuffer);
    OutputStream.Write(OutputBuffer, MimeEncodedSize(BytesRead));
  end;
end;

function MimeDecodedSize(const i: Integer): Integer;
begin
  Result := (i + 3) div 4 * 3;
end;

function MimeDecodePartialEnd(const OutputBuffer: Pointer; const ByteBuffer: Cardinal; const ByteBufferSpace: Cardinal): Integer;
var
  lByteBuffer: Cardinal;
begin
{$IFDEF RangeChecking}
  if OutputBuffer = nil then raise EMime.Create(SOutputBufferNil);
{$ENDIF}
  case ByteBufferSpace of
    1:
      begin
        lByteBuffer := ByteBuffer shr 2;
        PByte3(OutputBuffer).b2 := Byte(lByteBuffer);
        lByteBuffer := lByteBuffer shr 8;
        PByte3(OutputBuffer).b1 := Byte(lByteBuffer);
        Result := 2;
      end;
    2:
      begin
        lByteBuffer := ByteBuffer shr 4;
        PByte3(OutputBuffer).b1 := Byte(lByteBuffer);
        Result := 1;
      end;
  else
    Result := 0;
  end;
end;

function MimeDecodePartial(const InputBuffer: Pointer; const InputBytesCount: Integer; const OutputBuffer: Pointer; var ByteBuffer: Cardinal; var ByteBufferSpace: Cardinal): Integer;
var
  lByteBuffer, lByteBufferSpace, c: Cardinal;
  pIn, PInLimit: ^Byte;
  POut: PByte3;
begin
{$IFDEF RangeChecking}
  if InputBuffer = nil then raise EMime.Create(SInputBufferNil);
  if OutputBuffer = nil then raise EMime.Create(SOutputBufferNil);
{$ENDIF}
  if InputBytesCount > 0 then
  begin
    pIn := InputBuffer;
    Integer(PInLimit) := Integer(pIn) + InputBytesCount;

    POut := OutputBuffer;

    lByteBuffer := ByteBuffer;
    lByteBufferSpace := ByteBufferSpace;

    while pIn <> PInLimit do
    begin
      c := MIME_DECODE_TABLE[pIn^]; // Read from InputBuffer.
      Inc(pIn);

      if c = $FF then Continue;

      lByteBuffer := lByteBuffer shl 6;
      lByteBuffer := lByteBuffer or c;
      Dec(lByteBufferSpace);

      if lByteBufferSpace <> 0 then Continue; // Read 4 bytes from InputBuffer?

      POut.b3 := Byte(lByteBuffer); // Write 3 bytes to OutputBuffer (in reverse order).
      lByteBuffer := lByteBuffer shr 8;
      POut.b2 := Byte(lByteBuffer);
      lByteBuffer := lByteBuffer shr 8;
      POut.b1 := Byte(lByteBuffer);
      lByteBuffer := 0;
      Inc(POut);

      lByteBufferSpace := 4;
    end;

    ByteBuffer := lByteBuffer;
    ByteBufferSpace := lByteBufferSpace;

    Result := Integer(POut) - Integer(OutputBuffer);
  end
  else
    Result := 0;
end;

procedure MimeDecodeStream(const InputStream: TStream; const OutputStream: TStream);
var
  ByteBuffer, ByteBufferSpace: Cardinal;
  InputBuffer: array [0..BUFFER_SIZE - 1] of Byte;
  OutputBuffer: array [0..(BUFFER_SIZE + 3) div 4 * 3 - 1] of Byte;
  BytesRead: Integer;
begin
  ByteBuffer := 0;
  ByteBufferSpace := 4;
  BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  while BytesRead > 0 do
  begin
    OutputStream.Write(OutputBuffer, MimeDecodePartial(@InputBuffer, BytesRead, @OutputBuffer, ByteBuffer, ByteBufferSpace));
    BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  end;
  OutputStream.Write(OutputBuffer, MimeDecodePartialEnd(@OutputBuffer, ByteBuffer, ByteBufferSpace));
end;

function StrToken(var SrcStr: WideString; Token: WideString; IncludeTokenText: Boolean = False): WideString;
var
  i: Integer;
  t: Boolean;
begin
  repeat
    t := True;
    for i := 1 to Length(Token) do
      if SrcStr = '' then
      begin
        Result := '';
        Exit;
      end
      else if SrcStr[1] = Token[i] then
      begin
        Delete(SrcStr, 1, 1);
        t := False;
      end;
  until t;

  Result := '';
  for i := 1 to Length(Token) do
    if Pos(Token[i], SrcStr) > 0 then
    begin
      if IncludeTokenText then
        Result := Copy(SrcStr, 1, Pos(Token[i], SrcStr))
      else
        Result := Copy(SrcStr, 1, Pos(Token[i], SrcStr)-1);
      Delete(SrcStr, 1, Pos(Token[i], SrcStr));
      Exit;
    end;

  Result := SrcStr;
  SrcStr := '';
end;


function ExStrToDateTime(const Value: WideString): TDateTime;
var
  TestStr, DateStr, TimeStr: WideString;
  D1, D2, D3, T1, T2, T3: Word;
begin
  Result := 0;

  if Trim(Value) = '' then
  begin
    Result := 0;
    Exit;
  end;

  D1 := 0;
  D2 := 0;
  D3 := 0;
  T1 := 0;
  T2 := 0;
  T3 := 0;

  TestStr := UpperCase(Trim(Value));
  if Pos(' ', TestStr) > 0 then
  begin
    DateStr := Trim(Copy(TestStr, 1, Pos(' ', TestStr)));
    TimeStr := Trim(Copy(TestStr, Pos(' ', TestStr)+1, MaxInt));

    D1 := StrToIntDef(StrToken(DateStr, '/'), 0);
    D2 := StrToIntDef(StrToken(DateStr, '/'), 0);
    D3 := StrToIntDef(StrToken(DateStr, '/'), 0);
    T1 := StrToIntDef(StrToken(TimeStr, ':'), 0);
    T2 := StrToIntDef(StrToken(TimeStr, ':'), 0);
    T3 := StrToIntDef(StrToken(TimeStr, ':'), 0);
  end
  else if Pos(':', TestStr) > 0 then
  begin
    TimeStr := Trim(TestStr);
    T1 := StrToIntDef(StrToken(TimeStr, ':'), 0);
    T2 := StrToIntDef(StrToken(TimeStr, ':'), 0);
    T3 := StrToIntDef(StrToken(TimeStr, ':'), 0);
  end
  else if Pos('/', TestStr) > 0 then
  begin
    DateStr := Trim(TestStr);
    D1 := StrToIntDef(StrToken(DateStr, '/'), 0);
    D2 := StrToIntDef(StrToken(DateStr, '/'), 0);
    D3 := StrToIntDef(StrToken(DateStr, '/'), 0);
  end;

  try
    if ((D1 > 0) and (D2 in [1..12]) and (D3 in [1..31])) and ((T1 > 0) or (T2 > 0) or (T3 > 0)) then
      Result := EncodeDateTime(D1, D2, D3, T1, T2, T3, 0)
    else if (D1 > 0) and (D2 in [1..12]) and (D3 in [1..31]) then
      Result := EncodeDate(D1, D2, D3)
    else if (T1 > 0) or (T2 > 0) or (T3 > 0) then
      Result := EncodeTime(T1, T2, T3, 0);
  except
  end;

  if Result = 0 then
    try
      Result := StrToDateTime(TestStr);
    except
      Result := 0;
    end;
end;

function ExStrToDate(const Value: WideString): TDateTime;
var
  D1, D2, D3: Word;
begin
  Result := ExStrToDateTime(Value);
  if Result <> 0 then
  begin
    DecodeDate(Result, D1, D2, D3);
    Result := EncodeDate(D1, D2, D3);
  end;
end;

function ExStrToTime(const Value: WideString): TDateTime;
var
  T1, T2, T3, T4: Word;
begin
  Result := ExStrToDateTime(Value);
  if Result <> 0 then
  begin
    DecodeTime(Result, T1, T2, T3, T4);
    Result := EncodeTime(T1, T2, T3, T4);
  end;
end;

function ExDateTimeToStr(const Value: TDateTime): WideString;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    Result := FormatDateTime('YYYY/MM/DD', Value) + ' ' + FormatDateTime('HH:NN:SS', Value);
  end;
end;

function ExDateToStr(const Value: TDateTime): WideString;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    Result := FormatDateTime('YYYY/MM/DD', Value);
  end;
end;

function ExTimeToStr(const Value: TDateTime): WideString;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    Result := FormatDateTime('HH:NN:SS', Value);
  end;
end;

function Base64ToStr(Base64: WideString): WideString;
var
  S1, S2: TStringStream;
begin
  S2 := TStringStream.Create(Trim(Base64));
  try
    S1 := TStringStream.Create('');
    try
      S2.Position := 0;
      MimeDecodeStream(S2, S1);
      Result := S1.DataString;
    finally
      S1.Free;
    end;
  finally
    S2.Free;
  end;
end;

function StrToBase64(Str: WideString): WideString;
var
  S1, S2: TStringStream;
begin
  S1 := TStringStream.Create(Str);
  try
    S2 := TStringStream.Create('');
    try
      S1.Position := 0;
      MimeEncodeStream(S1, S2);
      Result := S2.DataString;
    finally
      S2.Free;
    end;
  finally
    S1.Free;
  end;
end;

type
  THandledObject = class(TObject)
  protected
    FHandle: THandle;
  public
    destructor Destroy; override;
    property Handle: THandle read FHandle;
  end;

  TMutex = class(THandledObject)
  public
    constructor Create(const Name: string);
    function Get(TimeOut: Integer = -1): Boolean;
    function Release: Boolean;
  end;

{ THandledObject }

destructor THandledObject.Destroy;
begin
  if FHandle <> 0 then
    CloseHandle(FHandle);
end;

{ TMutex }

constructor TMutex.Create(const Name: string);
begin
  FHandle := CreateMutex(nil, False, PChar(Name));
  if FHandle = 0 then abort;
end;

function TMutex.Get(TimeOut: Integer = -1): Boolean;
begin
  Result := WaitForSingleObject(FHandle, TimeOut) = WAIT_OBJECT_0;
end;

function TMutex.Release: Boolean;
begin
  Result := ReleaseMutex(FHandle);
end;

{ TXmlIni }

constructor TXmlIni.Create(AFileName: WideString);
begin
  CoInitialize(nil);

  FList := TList.Create;
  FXmlEncoding := XI_XML_DefaultEncoding;
  FAutoSave := False;
  FFileName := '';
  FCaseSensitive := False;
  FCompressed := False;
  FRaiseException := False;

  try
    LoadFromFile(AFileName);
  except
  end;
  FFileName := AFileName;
end;

destructor TXmlIni.Destroy;
begin
  if FAutoSave then
    SaveToFile;
  Clear;
  FList.Free;
end;

function TXmlIni.Equals(Dest: TXmlIni): Boolean;
var
  i: Integer;
begin
  Result := False;

  for i := 0 to Count-1 do
  begin
    if not Dest.Exists(SectionByIndex[i].Name) then
      Exit;
    if not SectionByIndex[i].Equals(Dest[SectionByIndex[i].Name]) then
      Exit;
  end;

  Result := True;
end;

procedure TXmlIni.CopyFrom(Source: TXmlIni);
begin
  Clear;
  Append(Source);
end;

procedure TXmlIni.Append(Source: TXmlIni);
var
  i: Integer;
begin
  for i := 0 to Source.Count-1 do
    Add(Source.SectionByIndex[i].Name).Append(Source.SectionByIndex[i]);
  Collation;
end;

function TXmlIni.Add(SectionName: WideString): TXmlIniSection;
begin
  Result := InternalFind(SectionName);
  if Result = nil then
    Result := InternalAdd(SectionName);
end;

procedure TXmlIni.Delete(SectionName: WideString);
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
    if SectionByIndex[i].Name = SectionName then
    begin
      SectionByIndex[i].Free;
      FList.Delete(i);
      Break;
    end;
end;

function TXmlIni.Exists(SectionName: WideString; KeyName: WideString = ''): Boolean;
var
  Section: TXmlIniSection;
begin
  Result := False;

  Section := InternalFind(SectionName);
  if Section <> nil then
  begin
    if KeyName <> '' then
      Result := Section.Exists(KeyName)
    else
      Result := True;
  end;
end;

procedure TXmlIni.Clear;
var
  i: Integer;
begin
  FXmlEncoding := XI_XML_DefaultEncoding;
  
  for i := Count - 1 downto 0 do
  begin
    SectionByIndex[i].Free;
    FList.Delete(i);
  end;
end;

procedure TXmlIni.Collation;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
  begin
    SectionByIndex[i].Collation;
    if SectionByIndex[i].Count = 0 then
    begin
      SectionByIndex[i].Free;
      FList.Delete(i);
    end;
  end;
end;

procedure TXmlIni.LoadFromStream(Stream: TStream);
var
  TestBuf: array[0..1] of Char;
  UnzipMS: TMemoryStream;
begin
  try
    Stream.Position := 0;
    Stream.Read(TestBuf, 2);

    if TestBuf = 'ZL' then
    begin
      FCompressed := True;
      Stream.Position := 0;
      UnzipMS := TMemoryStream.Create;
      try
        DecompressStream(Stream, UnzipMS);
        InternalXmlLoadFromStream(UnzipMS);
      finally
        UnzipMS.Free;
      end;
    end
    else
    begin
      FCompressed := False;
      Stream.Position := 0;
      InternalXmlLoadFromStream(Stream);
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIni.SaveToStream(Stream: TStream);
var
  UnzipMS: TMemoryStream;
begin
  try
    if FCompressed then
    begin
      UnzipMS := TMemoryStream.Create;
      try
        InternalXmlSaveToStream(UnzipMS);
        UnzipMS.Position := 0;
        Stream.Position := 0;
        CompressStream(UnzipMS, Stream);
      finally
        UnzipMS.Free;
      end;
    end
    else
      InternalXmlSaveToStream(Stream);
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIni.LoadFromString(Value: WideString);
var
  SS: TStringStream;
begin
  if TrimLeft(Value) = '' then
    Clear
  else if TrimLeft(Value)[1] = '<' then
  begin
    SS := TStringStream.Create(Value);
    try
      LoadFromStream(SS);
    finally
      SS.Free;
    end;
  end
  else
    LoadFromBase64String(Value);
end;

function TXmlIni.SaveToString: WideString;
var
  OldXIEncoding: TXmlIniEncodingType;
  MS: TMemoryStream;
  NullChar: WideChar;
begin
  if not FCompressed then
  begin
    OldXIEncoding := XmlEncoding;
    try
      XmlEncoding := xietUTF_16;

      MS := TMemoryStream.Create;
      try
        SaveToStream(MS);
        NullChar := #0;
        MS.Write(NullChar, 2);
        Result := PWideChar(PChar(MS.Memory)+2);
      finally
        MS.Free;
      end;
    finally
      XmlEncoding := OldXIEncoding;
    end;
  end
  else
    Result := SaveToBase64String;
end;

procedure TXmlIni.LoadFromBase64String(Value: string);
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  if (TrimLeft(Value) = '') or (TrimLeft(Value)[1] = '<') then
    LoadFromString(Value)
  else
  begin
    SS := TStringStream.Create(Trim(Value));
    try
      MS := TMemoryStream.Create;
      try
        SS.Position := 0;
        MimeDecodeStream(SS, MS);
        LoadFromStream(MS);
      finally
        MS.Free;
      end;
    finally
      SS.Free;
    end;
  end;
end;

function TXmlIni.SaveToBase64String: string;
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  MS := TMemoryStream.Create;
  try
    SaveToStream(MS);

    SS := TStringStream.Create('');
    try
      MS.Position := 0;
      MimeEncodeStream(MS, SS);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  finally
    MS.Free;
  end;
end;

procedure TXmlIni.LoadFromFile(AFileName: WideString = '');
begin
  AFileName := Trim(AFileName);
  
  if (AFileName <> '') and (ExtractFilePath(AFileName) = '') then
    AFileName := ExtractFilePath(ParamStr(0)) + AFileName;
    
  if AFileName = '' then
    AFileName := FFileName;
    
  if AFileName = '' then
    Exit;
    
  if not FileExists(AFileName) then
    if FRaiseException then
      raise Exception.Create('File "' + AFileName + '" not exists.')
    else
      Exit;

  FFileName := AFileName;

  InternalLoadFromFile(FFileName);
end;

procedure TXmlIni.SaveToFile(AFileName: WideString = '');
begin
  AFileName := Trim(AFileName);

  if (AFileName <> '') and (ExtractFilePath(AFileName) = '') then
    AFileName := ExtractFilePath(ParamStr(0)) + AFileName;

  if AFileName = '' then
    AFileName := FFileName;

  if AFileName = '' then
    if FRaiseException then
      raise Exception.Create('File name is required.')
    else
      Exit;

  FFileName := AFileName;

  InternalSaveToFile(FFileName);
end;

procedure TXmlIni.LoadFromRegistry(RootKey: HKEY; Path: WideString);
var
  Reg: TRegistry;
  Names: TStringList;
  i: Integer;
begin
  Clear;

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;

    if Reg.OpenKey(Path, False) then
    begin
      Names := TStringList.Create;
      try
        Reg.GetKeyNames(Names);
        for i := 0 to Names.Count-1 do
          Section[Names[i]].LoadFromRegistry(RootKey, Path, Names[i]);

        Section[''].LoadFromRegistry(RootKey, Path);

        Collation;
      finally
        Names.Free;
      end;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TXmlIni.SaveToRegistry(RootKey: HKEY; Path: WideString);
var
  i: Integer;
begin
  Collation;

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  for i := 0 to Count-1 do
    SectionByIndex[i].SaveToRegistry(RootKey, Path, SectionByIndex[i].Name);
end;

procedure TXmlIni.LoadFromIni(AIniFileName: WideString);
  function SingleLineToText(Str: WideString): WideString;
  begin
    Result := Str;
    
    if Pos('&', Result) > 0 then
    begin
      Result := StringReplace(Result, '&#xD;', #13, [rfIgnoreCase, rfReplaceAll]);
      Result := StringReplace(Result, '&#xA;', #10, [rfIgnoreCase, rfReplaceAll]);
      Result := StringReplace(Result, '&', '&amp;', [rfIgnoreCase, rfReplaceAll]);
    end;
  end;
var
  Ini: TIniFile;
  SectionNames, KeyNames: TStringList;
  SecName, KeyName: WideString;
  i, j: Integer;
begin
  Clear;

  AIniFileName := Trim(AIniFileName);

  if (AIniFileName <> '') and (ExtractFilePath(AIniFileName) = '') then
    AIniFileName := ExtractFilePath(ParamStr(0)) + AIniFileName;

  if (AIniFileName = '') or not FileExists(AIniFileName) then
    if FRaiseException then
      raise Exception.Create('File "' + AIniFileName + '" not exists.')
    else
      Exit;

  Ini := TIniFile.Create(AIniFileName);
  try
    SectionNames := TStringList.Create;
    KeyNames := TStringList.Create;
    try
      Ini.ReadSections(SectionNames);
      for i := 0 to SectionNames.Count-1 do
      begin
        SecName := SingleLineToText(SectionNames[i]);
        
        Ini.ReadSectionValues(SectionNames[i], KeyNames);
        for j := 0 to KeyNames.Count-1 do
        begin
          KeyName := SingleLineToText(KeyNames.Names[j]);
          Section[SecName][KeyName] := SingleLineToText(Ini.ReadString(SectionNames[i], KeyNames.Names[j], ''));
        end;
      end;

      Collation;
    finally
      KeyNames.Free;
      SectionNames.Free;
    end;
  finally
    Ini.Free;
  end;
end;

procedure TXmlIni.SaveToIni(AIniFileName: WideString);
  function TextToSingleLine(Str: WideString): WideString;
  begin
    Result := Str;
    if (Pos(#13, Result) > 0) or (Pos(#10, Result) > 0) then
    begin
      Result := StringReplace(Result, '&', '&amp;', [rfIgnoreCase, rfReplaceAll]);
      Result := StringReplace(Result, #10, '&#xA;', [rfIgnoreCase, rfReplaceAll]);
      Result := StringReplace(Result, #13, '&#xD;', [rfIgnoreCase, rfReplaceAll]);
    end;
  end;
var
  Ini: TIniFile;
  i, j: Integer;
begin
  Collation;

  AIniFileName := Trim(AIniFileName);

  if (AIniFileName <> '') and (ExtractFilePath(AIniFileName) = '') then
    AIniFileName := ExtractFilePath(ParamStr(0)) + AIniFileName;

  if AIniFileName = '' then
    if FRaiseException then
      raise Exception.Create('File name is required.')
    else
      Exit;

  DeleteFile(AIniFileName);

  Ini := TIniFile.Create(AIniFileName);
  try
    for i := 0 to Count-1 do
      for j := 0 to SectionByIndex[i].Count-1 do
        Ini.WriteString(
            TextToSingleLine(SectionByIndex[i].Name),
            TextToSingleLine(SectionByIndex[i].KeyByIndex[j].Name),
            TextToSingleLine(SectionByIndex[i].KeyByIndex[j].Text));
  finally
    Ini.Free;
  end;
end;

type
  TSetMemMemoryStream = class(TMemoryStream)
  public
    procedure SetMem(Ptr: Pointer; Size: Longint);
  end;

procedure TSetMemMemoryStream.SetMem(Ptr: Pointer; Size: Integer);
begin
  SetPointer(Ptr, Size);
end;

procedure TXmlIni.LoadFromShareMem(ShareMemID: string = '');
var
  MemLockMutex: TMutex;
  MS: TSetMemMemoryStream;
  ShareMemHandle: THandle;
  ShareMemMap: Pointer;
  ShareMemSize: Integer;
begin
  ShareMemID := Trim(ShareMemID);
  
  if ShareMemID = '' then
    Exit;

  ShareMemID := XI_ShareName + ShareMemID; 

  MemLockMutex := TMutex.Create(XI_LockMutexName + ShareMemID);
  try
    if MemLockMutex.Get then
    begin
      try
        try
          ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READONLY, 0,
              SizeOf(Integer), PChar(ShareMemID));

          ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_READ, 0, 0, SizeOf(Integer));
          try
            Move(ShareMemMap^, ShareMemSize, SizeOf(Integer));
          finally
            UnmapViewOfFile(ShareMemMap);
          end;

          if ShareMemSize = 0 then
            Clear
          else
          begin
            ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READONLY, 0,
                SizeOf(Integer) + ShareMemSize, PChar(ShareMemID));

            ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_READ, 0, 0, SizeOf(Integer) + ShareMemSize);
            try
              MS := TSetMemMemoryStream.Create;
              try
                MS.SetMem(Pointer(PChar(ShareMemMap) + SizeOf(Integer)), ShareMemSize);
                LoadFromStream(MS);
                MS.SetMem(nil, 0);
              finally
                MS.Free;
              end;
            finally
              UnmapViewOfFile(ShareMemMap);
            end;
          end;
        except
        end;
      finally
        MemLockMutex.Release;
      end;
    end;
  finally
    MemLockMutex.Free;
  end;
end;

procedure TXmlIni.SaveToShareMem(ShareMemID: string = '');
var
  MemLockMutex: TMutex;
  MS: TMemoryStream;
  ShareMemHandle: THandle;
  ShareMemMap: Pointer;
  ShareMemSize: Integer;
begin
  ShareMemID := Trim(ShareMemID);

  if ShareMemID = '' then
    if FRaiseException then
      raise Exception.Create('ShareMemID is required.')
    else
      Exit;

  ShareMemID := XI_ShareName + ShareMemID;

  MemLockMutex := TMutex.Create(XI_LockMutexName + ShareMemID);
  try
    if MemLockMutex.Get then
    begin
      try
        try
          MS := TMemoryStream.Create;
          try
            SaveToStream(MS);

            ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READWRITE, 0,
                SizeOf(Integer) + MS.Size, PChar(ShareMemID));

            ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_WRITE, 0, 0, SizeOf(Integer) + MS.Size);
            try
              ShareMemSize := MS.Size;
              Move(ShareMemSize, ShareMemMap^, SizeOf(Integer));
              Move(MS.Memory^, Pointer(PChar(ShareMemMap) + SizeOf(Integer))^, MS.Size);
            finally
              UnmapViewOfFile(ShareMemMap);
            end;
          finally
            MS.Free;
          end;
        except
        end;
      finally
        MemLockMutex.Release;
      end;
    end;
  finally
    MemLockMutex.Free;
  end;
end;

function TXmlIni.SectionExists(const Section: WideString): Boolean;
begin
  Result := Exists(Section);
end;

function TXmlIni.ReadString(const Section, Ident, Default: WideString): WideString;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section][Ident]
  else
    Result := Default;
end;

procedure TXmlIni.WriteString(const Section, Ident, Value: WideString);
begin
  Self.Section[Section][Ident] := Value;
end;

function TXmlIni.ReadInteger(const Section, Ident: WideString; Default: Longint): Longint;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsInteger
  else
    Result := Default;
end;

procedure TXmlIni.WriteInteger(const Section, Ident: WideString; Value: Longint);
begin
  Self.Section[Section].Key[Ident].AsInteger := Value;
end;

function TXmlIni.ReadFloat(const Section, Ident: WideString; Default: Double): Double;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsFloat
  else
    Result := Default;
end;

procedure TXmlIni.WriteFloat(const Section, Ident: WideString; Value: Double);
begin
  Self.Section[Section].Key[Ident].AsFloat := Value;
end;

function TXmlIni.ReadBool(const Section, Ident: WideString; Default: Boolean): Boolean;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsBool
  else
    Result := Default;
end;

procedure TXmlIni.WriteBool(const Section, Ident: WideString; Value: Boolean);
begin
  Self.Section[Section].Key[Ident].AsBool := Value;
end;

function TXmlIni.ReadDateTime(const Section, Ident: WideString; Default: TDateTime): TDateTime;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsDateTime
  else
    Result := Default;
end;

procedure TXmlIni.WriteDateTime(const Section, Ident: WideString; Value: TDateTime);
begin
  Self.Section[Section].Key[Ident].AsDateTime := Value;
end;

function TXmlIni.ReadDate(const Section, Ident: WideString; Default: TDateTime): TDateTime;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsDate
  else
    Result := Default;
end;

procedure TXmlIni.WriteDate(const Section, Ident: WideString; Value: TDateTime);
begin
  Self.Section[Section].Key[Ident].AsDate := Value;
end;

function TXmlIni.ReadTime(const Section, Ident: WideString; Default: TDateTime): TDateTime;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsTime
  else
    Result := Default;
end;

procedure TXmlIni.WriteTime(const Section, Ident: WideString; Value: TDateTime);
begin
  Self.Section[Section].Key[Ident].AsTime := Value;
end;

function TXmlIni.ReadBinaryStream(const Section, Ident: WideString; Value: TStream): Integer;
begin
  Result := Self.Section[Section].Key[Ident].AsStream.Size;
  if Result > 0 then
    Value.CopyFrom(Self.Section[Section].Key[Ident].AsStream, 0);
end;

procedure TXmlIni.WriteBinaryStream(const Section, Ident: WideString; Value: TStream);
begin
  Self.Section[Section].Key[Ident].AsStream := Value;
end;

procedure TXmlIni.ReadSection(const Section: WideString; Strings: TStrings);
begin
  ReadSectionValues(Section, Strings);
end;

procedure TXmlIni.ReadSections(Strings: TStrings);
var
  i: Integer;
begin
  Strings.BeginUpdate;
  try
    Strings.Clear;
    for i := 0 to Count-1 do
      Strings.Add(SectionByIndex[i].Name);
  finally
    Strings.EndUpdate;
  end;
end;

procedure TXmlIni.ReadSectionValues(const Section: WideString; Strings: TStrings);
var
  i: Integer;
begin
  Strings.BeginUpdate;
  try
    Strings.Clear;
    if Exists(Section) then
      for i := 0 to Self.Section[Section].Count-1 do
        Strings.Add(Self.Section[Section].KeyByIndex[i].Name + '=' + Self.Section[Section].KeyByIndex[i].Text);
  finally
    Strings.EndUpdate;
  end;
end;

procedure TXmlIni.EraseSection(const Section: WideString);
begin
  Self.Delete(Section);
end;

procedure TXmlIni.DeleteKey(const Section, Ident: WideString);
begin
  if SectionExists(Section) then
    Self.Section[Section].Key[Ident].Clear;
end;

procedure TXmlIni.UpdateFile;
begin
  SaveToFile;
end;

function TXmlIni.ValueExists(const Section, Ident: WideString): Boolean;
begin
  if SectionExists(Section) then
    Result := Self.Section[Section].Exists(Ident)
  else
    Result := False;
end;

procedure TXmlIni.InternalXmlLoadFromStream(Stream: TStream);
var
  XML: TXMLDocument;
  Root: IXmlNode;
  Enc: TXmlIniEncodingType;
  i: Integer;
  j: Integer;
begin
  Clear;

  XML := TXMLDocument.Create(Application);
  try
    Stream.Position := 0;
    XML.LoadFromStream(Stream, xetUnknown);

    FXmlEncoding := xietUnknow;

    for Enc := Low(XI_XML_EncodingStr) to High(XI_XML_EncodingStr) do
      if XML.Encoding = XI_XML_EncodingStr[Enc] then
      begin
        FXmlEncoding := Enc;
        Break;
      end;

    Root := XML.DocumentElement;

    if Root.NodeName <> XI_Root_NodeName then
      Exit;

    for i := 0 to Root.ChildNodes.Count - 1 do
      if Root.ChildNodes[i].NodeName = XI_Section_NodeName then
        with Add(Root.ChildNodes[i].Attributes[XI_Name_Attribute]) do
        begin
          for j := 0 to Root.ChildNodes[i].ChildNodes.Count - 1 do
            if Root.ChildNodes[i].ChildNodes[j].NodeName = XI_Key_NodeName then
              Item[Root.ChildNodes[i].ChildNodes[j].Attributes[XI_Name_Attribute]] :=
                Root.ChildNodes[i].ChildNodes[j].Text;
        end;

    Collation;
  finally
    XML.Free;
  end;
end;

procedure TXmlIni.InternalXmlSaveToStream(Stream: TStream);
  function GetLocalCodePage: Integer;
  const
    CP_ACP = 0;
    LOCALE_IDEFAULTANSICODEPAGE = $00001004;
  var
    ResultCode: Integer;
    Buffer: array [0..6] of Char;
  begin
    GetLocaleInfo(GetThreadLocale, LOCALE_IDEFAULTANSICODEPAGE, Buffer, SizeOf(Buffer));
    Val(Buffer, Result, ResultCode);
    if ResultCode <> 0 then
      Result := CP_ACP;
  end;
  function GetLocalEncodingType: TXmlIniEncodingType;
  begin
    case GetLocalCodePage of
      950:  Result := xietBIG5;
      936:  Result := xietGB2312;
      932:  Result := xietS_JIS;
      949:  Result := xietEUC_KR;
    else
      Result := xietUTF_8;
    end;
  end;
var
  XML: TXMLDocument;
  Root, SectionNode, KeyNode: IXmlNode;
  i: Integer;
  j: Integer;
begin
  Collation;

  if FXmlEncoding = xietUnknow then
    FXmlEncoding := GetLocalEncodingType;

  XML := TXMLDocument.Create(Application);
  try
    XML.Active := True;
    XML.Options := XML.Options + [doNodeAutoIndent];
    XML.NodeIndentStr := '  ';

    Root := XML.AddChild(XI_ROOT_NodeName);

    for i := 0 to Count - 1 do
    begin
      SectionNode := Root.AddChild(XI_Section_NodeName);
      SectionNode.Attributes[XI_Name_Attribute] := SectionByIndex[i].Name;
      for j := 0 to SectionByIndex[i].Count - 1 do
      begin
        KeyNode := SectionNode.AddChild(XI_Key_NodeName);
        KeyNode.Attributes[XI_Name_Attribute] := SectionByIndex[i].KeyByIndex[j].Name;
        KeyNode.Text := SectionByIndex[i].KeyByIndex[j].Text;
      end;
    end;

    XML.StandAlone := XI_XML_StandAlone;
    XML.Version := XI_XML_Version;
    XML.Encoding := XI_XML_EncodingStr[FXmlEncoding];

    Stream.Position := 0;
    if Stream is TStringStream then
      TStringStream(Stream).WriteString(XML.XML.Text)
    else
      XML.SaveToStream(Stream);
  finally
    XML.Free;
  end;
end;

procedure TXmlIni.InternalLoadFromFile(AFileName: WideString);
var
  FS: TFileStream;
  FileLockMutex: TMutex;
begin
  Clear;

  FileLockMutex := TMutex.Create(
      StringReplace(XI_LockMutexName + AFileName, '\', '_', [rfReplaceAll, rfIgnoreCase]));
  try
    if FileLockMutex.Get then
    begin
      try
        FS := TFileStream.Create(AFileName, fmOpenRead and fmShareDenyWrite);
        try
          LoadFromStream(FS);
        finally
          FS.Free;
        end;
      finally
        FileLockMutex.Release;
      end;
    end;
  finally
    FileLockMutex.Free;
  end;
end;

procedure TXmlIni.InternalSaveToFile(AFileName: WideString);
var
  FS: TFileStream;
  FileLockMutex: TMutex;
begin
  FileLockMutex := TMutex.Create(
      StringReplace(XI_LockMutexName + AFileName, '\', '_', [rfReplaceAll, rfIgnoreCase]));
  try
    if FileLockMutex.Get then
    begin
      try
        FS := TFileStream.Create(AFileName, fmCreate);
        try
          SaveToStream(FS);
        finally
          FS.Free;
        end;
      finally
        FileLockMutex.Release;
      end;
    end;
  finally
    FileLockMutex.Free;
  end;
end;

function TXmlIni.InternalFind(SectionName: WideString): TXmlIniSection;
var
  i: Integer;
begin
  Result := nil;

  SectionName := Trim(SectionName);

  for i := Count - 1 downto 0 do
    if (SectionByIndex[i].Name = SectionName) or
      (not CaseSensitive and (UpperCase(SectionByIndex[i].Name) = UpperCase(SectionName))) then
    begin
      Result := SectionByIndex[i];
      Exit;
    end;
end;

function TXmlIni.InternalAdd(SectionName: WideString): TXmlIniSection;
begin
  Result := TXmlIniSection.Create(SectionName);
  Result.CaseSensitive := FCaseSensitive;
  FList.Add(Result);
end;

function TXmlIni.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TXmlIni.GetSectionByIndex(Index: Integer): TXmlIniSection;
begin
  if (Index < 0) or (Index >= FList.Count) then
    Result := GetSection('')
  else
    Result := TXmlIniSection(FList[Index]);
end;

function TXmlIni.GetSection(Index: WideString): TXmlIniSection;
begin
  Result := Add(Index);
end;

procedure TXmlIni.SetFileName(const Value: WideString);
begin
  if FFileName <> Value then
  begin
    if FAutoSave then
      SaveToFile;
    FFileName := Value;
    LoadFromFile;
  end;
end;

procedure TXmlIni.SetAutoSave(const Value: Boolean);
begin
  FAutoSave := Value;
  if FAutoSave then
    SaveToFile;
end;

procedure TXmlIni.SetCaseSensitive(const Value: Boolean);
var
  i: Integer;
begin
  if FCaseSensitive <> Value then
  begin
    FCaseSensitive := Value;

    for i := 0 to Count-1 do
      SectionByIndex[i].CaseSensitive := FCaseSensitive;
  end;
end;

procedure TXmlIni.SetCompressed(const Value: Boolean);
var
  i: Integer;
begin
  if FCompressed <> Value then
  begin
    FCompressed := Value;

    if FAutoSave then
      SaveToFile;

    for i := 0 to Count-1 do
      SectionByIndex[i].Compressed := FCompressed;
  end;
end;

procedure TXmlIni.SetRaiseException(const Value: Boolean);
var
  i: Integer;
begin
  if FRaiseException <> Value then
  begin
    FRaiseException := Value;
    for i := 0 to Count-1 do
      SectionByIndex[i].RaiseException := FRaiseException;
  end;
end;

{ TXmlIniSection }

constructor TXmlIniSection.Create(AName: WideString = ''; AOwner: TXmlIni = nil);
begin
  FName := Trim(AName);
  FOwner := AOwner;
  FList := TList.Create;
  
  FXmlEncoding := XI_XML_DefaultEncoding;
  FCaseSensitive := False;
  FCompressed := False;
  if AOwner <> nil then
    FRaiseException := FOwner.RaiseException
  else
    FRaiseException := False;
end;

destructor TXmlIniSection.Destroy;
begin
  Clear;
  FList.Free;

  inherited;
end;

function TXmlIniSection.Equals(Dest: TXmlIniSection): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Count-1 do
  begin
    if not Dest.Exists(KeyByIndex[i].Name) then
      Exit;
    if Dest[KeyByIndex[i].Name] <> KeyByIndex[i].Text then
      Exit;
  end;
  Result := True;
end;

procedure TXmlIniSection.CopyFrom(Source: TXmlIniSection);
begin
  Clear;
  Append(Source);
end;

procedure TXmlIniSection.Append(Source: TXmlIniSection);
var
  i: Integer;
begin
  for i := 0 to Source.Count-1 do
    SetItem(Source.KeyByIndex[i].Name, Source.KeyByIndex[i].Text);
end;

function TXmlIniSection.Equals(Dest: TStrings): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Count-1 do
  begin
    if Dest.IndexOfName(KeyByIndex[i].Name) < 0 then
      Exit;
    if Dest.Values[KeyByIndex[i].Name] <> KeyByIndex[i].Text then
      Exit;
  end;
  Result := True;
end;

procedure TXmlIniSection.CopyFrom(Source: TStrings);
begin
  Clear;
  Append(Source);
end;

procedure TXmlIniSection.Append(Source: TStrings);
var
  i: Integer;
begin
  for i := 0 to Source.Count-1 do
    SetItem(Source.Names[i], Source.Values[Source.Names[i]]);
end;

function TXmlIniSection.Exists(KeyName: WideString): Boolean;
begin
  Result := InternalFind(KeyName) <> nil;
end;

procedure TXmlIniSection.Clear;
var
  i: Integer;
begin
  FXmlEncoding := XI_XML_DefaultEncoding;
  
  for i := Count - 1 downto 0 do
  begin
    KeyByIndex[i].Free;
    FList.Delete(i);
  end;
end;

procedure TXmlIniSection.Collation;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
    if KeyByIndex[i].Text = '' then
    begin
      KeyByIndex[i].Free;
      FList.Delete(i);
    end;
end;

procedure TXmlIniSection.LoadFromStream(Stream: TStream; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  XmlIni := TXmlIni.Create;
  try
    XmlIni.LoadFromStream(Stream);

    if (SectionName = '') and not XmlIni.Exists('') then
    begin
      if XmlIni.Exists(FName) then
        SectionName := FName
      else if XmlIni.Count > 0 then
        SectionName := XmlIni.SectionByIndex[0].Name;
    end;

    CopyFrom(XmlIni[SectionName]);
    FName := XmlIni[SectionName].Name;
    
    FXmlEncoding := XmlIni.XmlEncoding;
    FCompressed := XmlIni.Compressed;
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.SaveToStream(Stream: TStream; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  if SectionName = '' then
    SectionName := FName;

  XmlIni := TXmlIni.Create;
  try
    XmlIni[SectionName].CopyFrom(Self);
    XmlIni.XmlEncoding := FXmlEncoding;
    XmlIni.Compressed := FCompressed;
    XmlIni.SaveToStream(Stream);
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.LoadFromString(Value: WideString; SectionName: WideString = '');
var
  SS: TStringStream;
begin
  if TrimLeft(Value) = '' then
    Clear
  else if TrimLeft(Value)[1] = '<' then
  begin
    SS := TStringStream.Create(Value);
    try
      LoadFromStream(SS, SectionName);
    finally
      SS.Free;
    end;
  end
  else
    LoadFromBase64String(Value, SectionName);
end;

function TXmlIniSection.SaveToString(SectionName: WideString = ''): WideString;
var
  SS: TStringStream;
begin
  if not FCompressed then
  begin
    SS := TStringStream.Create('');
    try
      SaveToStream(SS, SectionName);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  end
  else
    Result := SaveToBase64String;
end;

procedure TXmlIniSection.LoadFromBase64String(Value: string; SectionName: WideString = '');
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  if (TrimLeft(Value) = '') or (TrimLeft(Value)[1] = '<') then
    LoadFromString(Value)
  else
  begin
    SS := TStringStream.Create(Trim(Value));
    try
      MS := TMemoryStream.Create;
      try
        SS.Position := 0;
        MimeDecodeStream(SS, MS);
        LoadFromStream(MS, SectionName);
      finally
        MS.Free;
      end;
    finally
      SS.Free;
    end;
  end;
end;

function TXmlIniSection.SaveToBase64String(SectionName: WideString = ''): string;
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  MS := TMemoryStream.Create;
  try
    SaveToStream(MS, SectionName);

    SS := TStringStream.Create('');
    try
      MS.Position := 0;
      MimeEncodeStream(MS, SS);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  finally
    MS.Free;
  end;
end;

procedure TXmlIniSection.LoadFromFile(FileName: WideString; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  XmlIni := TXmlIni.Create;
  try
    XmlIni.LoadFromFile(FileName);

    if (SectionName = '') and not XmlIni.Exists('') then
    begin
      if XmlIni.Exists(FName) then
        SectionName := FName
      else if XmlIni.Count > 0 then
        SectionName := XmlIni.SectionByIndex[0].Name;
    end;

    CopyFrom(XmlIni[SectionName]);
    FName := XmlIni[SectionName].Name;

    FXmlEncoding := XmlIni.XmlEncoding;
    FCompressed := XmlIni.Compressed;
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.SaveToFile(FileName: WideString; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  if SectionName = '' then
    SectionName := FName;

  XmlIni := TXmlIni.Create;
  try
    XmlIni[SectionName].CopyFrom(Self);
    XmlIni.XmlEncoding := FXmlEncoding;
    XmlIni.Compressed := FCompressed;
    XmlIni.SaveToFile(FileName);
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.LoadFromRegistry(RootKey: HKEY; Path: WideString; SectionName: WideString = '');
var
  Reg: TRegistry;
  RegDataInfo: TRegDataInfo;
  Names: TStringList;
  i: Integer;
  Buf: PChar;
begin
  Clear;

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;

    if SectionName <> '' then
      Path := Path + SectionName;

    Names := TStringList.Create;
    try
      if Reg.OpenKey(Path, False) then
      begin
        Reg.GetValueNames(Names);

        if Names.Count = 0 then
        begin
          Reg.CloseKey;
          if Reg.OpenKey(Path, False) then
            Reg.GetValueNames(Names);
        end;

        for i := 0 to Names.Count-1 do
        begin
          Reg.GetDataInfo(Names[i], RegDataInfo);

          case RegDataInfo.RegData of
            rdString, rdExpandString:
              Key[Names[i]].Text := Reg.ReadString(Names[i]);
            rdInteger:
              Key[Names[i]].AsInteger := Reg.ReadInteger(Names[i]);
            rdUnknown, rdBinary:
              begin
                Buf := AllocMem(RegDataInfo.DataSize);
                try
                  Reg.ReadBinaryData(Names[i], Buf^, RegDataInfo.DataSize);
                  Key[Names[i]].FStream.Write(Buf^, RegDataInfo.DataSize);
                finally
                  FreeMem(Buf);
                end;
              end;
          end;
        end;

        Reg.CloseKey;
      end;

      Collation;
    finally
      Names.Free;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TXmlIniSection.SaveToRegistry(RootKey: HKEY; Path: WideString; SectionName: WideString = '';
    UseOldDataType: Boolean = False);
var
  Reg: TRegistry;
  RegDataInfo: TRegDataInfo;
  Names: TStringList;
  i: Integer;
begin
  Collation;

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;

    if SectionName <> '' then
      Path := Path + SectionName;

    Names := TStringList.Create;
    try
      if Reg.OpenKey(Path, True) then
      begin
        Reg.GetValueNames(Names);

        for i := 0 to Names.Count-1 do
          if not UseOldDataType or not Exists(Names[i]) then
            Reg.DeleteValue(Names[i]);

        for i := 0 to Count-1 do
        begin
          if UseOldDataType and Reg.ValueExists(KeyByIndex[i].Name) then
          begin
            Reg.GetDataInfo(KeyByIndex[i].Name, RegDataInfo);

            case RegDataInfo.RegData of
              rdString, rdExpandString:
                Reg.WriteString(KeyByIndex[i].Name, KeyByIndex[i].Text);
              rdInteger:
                Reg.WriteInteger(KeyByIndex[i].Name, KeyByIndex[i].AsInteger);
              rdUnknown, rdBinary:
                begin
                  KeyByIndex[i].TextToBinary;
                  Reg.WriteBinaryData(KeyByIndex[i].Name, KeyByIndex[i].FStream.Memory^, KeyByIndex[i].FStream.Size);
                end;
              else
            end;
          end
          else
          begin
            case KeyByIndex[i].DataType of
              xidtInteger:
                Reg.WriteInteger(KeyByIndex[i].Name, KeyByIndex[i].AsInteger);
              xidtText, xidtDateTime, xidtBool:
                Reg.WriteString(KeyByIndex[i].Name, KeyByIndex[i].Text);
              xidtBinary:
                begin
                  KeyByIndex[i].TextToBinary;
                  Reg.WriteBinaryData(KeyByIndex[i].Name, KeyByIndex[i].FStream.Memory^, KeyByIndex[i].FStream.Size);
                end;
            end;
          end;
        end;

        Reg.CloseKey;
      end;
    finally
      Names.Free;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TXmlIniSection.LoadFromIni(FileName, SectionName: WideString);
var
  XmlIni: TXmlIni;
begin
  XmlIni := TXmlIni.Create;
  try
    XmlIni.LoadFromIni(FileName);

    if (SectionName = '') and not XmlIni.Exists('') then
    begin
      if XmlIni.Exists(FName) then
        SectionName := FName
      else if XmlIni.Count > 0 then
        SectionName := XmlIni.SectionByIndex[0].Name;
    end;

    CopyFrom(XmlIni[SectionName]);
    FName := XmlIni[SectionName].Name;

    FXmlEncoding := XmlIni.XmlEncoding;
    FCompressed := XmlIni.Compressed;
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.SaveToIni(FileName, SectionName: WideString);
var
  XmlIni: TXmlIni;
begin
  if SectionName = '' then
    SectionName := FName;

  XmlIni := TXmlIni.Create;
  try
    XmlIni[SectionName].CopyFrom(Self);
    XmlIni.XmlEncoding := FXmlEncoding;
    XmlIni.Compressed := FCompressed;
    XmlIni.SaveToFile(FileName);
  finally
    XmlIni.Free;
  end;
end;

function TXmlIniSection.InternalAdd(KeyName: WideString): TXmlIniKey;
begin
  Result := TXmlIniKey.Create(KeyName);
  FList.Add(Result);
end;

function TXmlIniSection.InternalFind(KeyName: WideString): TXmlIniKey;
var
  i: Integer;
begin
  Result := nil;

  KeyName := Trim(KeyName);

  for i := 0 to Count - 1 do
    if (KeyByIndex[i].Name = KeyName) or
      (not FCaseSensitive and (UpperCase(KeyByIndex[i].Name) = UpperCase(KeyName))) then
    begin
      Result := KeyByIndex[i];
      Break;
    end;
end;

procedure TXmlIniSection.SetName(const Value: WideString);
begin
  try
    if (FOwner <> nil) and not FOwner.Exists(Value) then
      raise Exception.Create('Section "' + Value + '" already exists.');
    FName := Value;
  except
    if FRaiseException then
      raise;
  end;
end;

function TXmlIniSection.GetKeyByIndex(Index: Integer): TXmlIniKey;
begin
  if (Index < 0) or (Index >= FList.Count) then
    Result := GetKey('') 
  else
    Result := TXmlIniKey(FList[Index]);
end;

function TXmlIniSection.GetKey(Index: WideString): TXmlIniKey;
var
  Key: TXmlIniKey;
begin
  Key := InternalFind(Index);
  if Key = nil then
    Key := InternalAdd(Index);
  Result := Key;
end;

function TXmlIniSection.GetItem(Index: WideString): WideString;
begin
  Result := Key[Index].Text;
end;

procedure TXmlIniSection.SetItem(Index: WideString; const Value: WideString);
begin
  Key[Index].Text := Value;
end;

function TXmlIniSection.GetCount: Integer;
begin
  Result := FList.Count;
end;

procedure TXmlIniSection.SetRaiseException(const Value: Boolean);
var
  i: Integer;
begin
  if FRaiseException <> Value then
  begin
    FRaiseException := Value;
    for i := 0 to Count-1 do
      KeyByIndex[i].RaiseException := FRaiseException;
  end;
end;

{ TXmlIniKey }

constructor TXmlIniKey.Create(AName: WideString = ''; AText: WideString = ''; AOwner: TXmlIniSection = nil);
begin
  FOwner := AOwner;
  FName := Trim(AName);
  FStream := TXmlIniEventMemoryStream.Create(Self);
  FDataType := xidtText;
  if AOwner <> nil then
    FRaiseException := FOwner.RaiseException
  else
    FRaiseException := False;
  Text := AText;
end;

destructor TXmlIniKey.Destroy;
begin
  FStream.Free;

  inherited;
end;

procedure TXmlIniKey.Clear;
begin
  FText := '';
  FStream.Clear;
  FDataType := xidtText;
end;

procedure TXmlIniKey.ValueLoadFromFile(FileName: WideString; Binary: Boolean = True);
var
  SS: TStringStream;
  FS: TFileStream;
begin
  try
    Clear;

    FileName := Trim(FileName);

    if (FileName <> '') and (ExtractFilePath(FileName) = '') then
      FileName := ExtractFilePath(ParamStr(0)) + FileName;

    if (FileName = '') or not FileExists(FileName) then
      if FRaiseException then
        raise Exception.Create('File "' + FileName + '" not exists.')
      else
        Exit;

    if Binary then
      FStream.LoadFromFile(FileName)
    else
    begin
      SS := TStringStream.Create('');
      try
        FS := TFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
        try
          SS.CopyFrom(FS, 0);
        finally
          FS.Free;
        end;
        Text := SS.DataString;
      finally
        SS.Free;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.ValueSaveToFile(FileName: WideString);
var
  SS: TStringStream;
  FS: TFileStream;
begin
  FileName := Trim(FileName);
  
  if (FileName <> '') and (ExtractFilePath(FileName) = '') then
    FileName := ExtractFilePath(ParamStr(0)) + FileName;
    
  if FileName = '' then
    if FRaiseException then
      raise Exception.Create('File name is required.')
    else
      Exit;

  try
    if FDataType = xidtBinary then
    begin
      if FStream.Size > 0 then
        FStream.SaveToFile(FileName);
    end
    else
    begin
      if FText <> '' then
      begin
        SS := TStringStream.Create(FText);
        try
          FS := TFileStream.Create(FileName, fmCreate);
          try
            FS.CopyFrom(SS, 0);
          finally
            FS.Free;
          end;
        finally
          SS.Free;
        end;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.SetName(const Value: WideString);
begin
  try
    if (FOwner <> nil) and not FOwner.Exists(Value) then
      raise Exception.Create('Key "' + Value + '" already exists.');
    FName := Value;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.BinaryToText;
var
  SS: TStringStream;
begin
  try
    if FDataType = xidtBinary then
    begin
      SS := TStringStream.Create('');
      try
        FStream.Position := 0;
        MimeEncodeStream(FStream, SS);
        FText := XI_Base64Header + SS.DataString;
      finally
        SS.Free;
      end;
    end
    else
    begin
      SS := TStringStream.Create('');
      try
        SS.CopyFrom(FStream, 0);
        FText := SS.DataString;
      finally
        SS.Free;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.TextToBinary;
var
  SS: TStringStream;
  Base64Text: WideString;
  OldDataType: TXmlIniDataType;
begin
  try
    if Copy(FText, 1, Length(XI_Base64Header)) <> XI_Base64Header then
    begin
      SS := TStringStream.Create(FText);
      try
        OldDataType := FDataType;
        FStream.CopyFrom(SS, 0);
        FDataType := OldDataType;
      finally
        SS.Free;
      end;
    end
    else
    begin
      Base64Text := Trim(FText);

      Delete(Base64Text, 1, 8);

      SS := TStringStream.Create(Base64Text);
      try
        MimeDecodeStream(SS, FStream);
      finally
        SS.Free;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

function TXmlIniKey.GetText: WideString;
begin
  if (FText = '') and (FStream.Size > 0) then
    BinaryToText;
    
  Result := FText;
end;

procedure TXmlIniKey.SetText(const Value: WideString);
  function InArray(Str: WideString; ArrayStr: array of WideString): Boolean;
  var
    i: Integer;
  begin
    Result := False;

    Str := UpperCase(Str);
    
    for i := Low(ArrayStr) to High(ArrayStr) do
      if Str = UpperCase(ArrayStr[i]) then
      begin
        Result := True;
        Break;
      end;
  end;
  function IsAllNumber(Str: WideString): Boolean;
  var
    i: Integer;
  begin
    Result := True;
    Str := Trim(Str);
    if Str <> '' then
      for i := 1 to Length(Str) do
        if not ((Str[i] >= '0') and (Str[i] < '9')) then
        begin
          Result := False;
          Break;
        end;
  end;
  function IsFloat(Str: WideString): Boolean;
  begin
    Str := Trim(Str);
    Result := Str = FloatToStr(StrToFloatDef(Str, 0));
  end;
  function IsDateTime(Str: WideString): Boolean;
  begin
    Str := Trim(Str);
    Result := Str = ExDateTimeToStr(ExStrToDateTime(Str));
  end;
  function IsDate(Str: WideString): Boolean;
  begin
    Str := Trim(Str);
    Result := Str = ExDateToStr(ExStrToDate(Str));
  end;
  function IsTime(Str: WideString): Boolean;
  begin
    Str := Trim(Str);
    Result := Str = ExDateToStr(ExStrToTime(Str));
  end;
begin
  if FText <> Value then
  begin
    FStream.Clear;

    FText := Value;
    
    if (FText <> '') and (Copy(FText, 1, Length(XI_Base64Header)) = XI_Base64Header) then
      FDataType := xidtBinary
    else if InArray(FText, XI_True_Text) or InArray(FText, XI_False_Text) then
      FDataType := xidtBool
    else if IsAllNumber(FText) then
      FDataType := xidtInteger
    else if IsFloat(FText) then
      FDataType := xidtFloat
    else if IsDateTime(FText) then
      FDataType := xidtDateTime
    else if IsDate(FText) then
      FDataType := xidtDate
    else if IsTime(FText) then
      FDataType := xidtTime
    else
      FDataType := xidtText;
  end;
end;

function TXmlIniKey.GetAsStream: TStream;
begin
  Result := FStream;
end;

procedure TXmlIniKey.SetAsStream(const Value: TStream);
begin
  FText := '';

  if (Value <> nil) and (Value.Size > 0) then
  begin
    Value.Position := 0;
    FStream.CopyFrom(Value, 0);
    FDataType := xidtBinary;
  end
  else
  begin
    FStream.Clear;
    FDataType := xidtText;
  end;
end;

function TXmlIniKey.GetAsInteger: Integer;
begin
  Result := StrToIntDef(Text, 0);
end;

procedure TXmlIniKey.SetAsInteger(const Value: Integer);
begin
  Text := IntToStr(Value);
  FDataType := xidtInteger;
end;

function TXmlIniKey.GetAsFloat: Double;
begin
  Result := StrToFloatDef(Text, 0);
end;

procedure TXmlIniKey.SetAsFloat(const Value: Double);
begin
  Text := FloatToStr(Value);
  FDataType := xidtFloat;
end;

function TXmlIniKey.GetAsBool: Boolean;
var
  TestStr: WideString;
  i: Integer;
begin
  TestStr := UpperCase(Trim(Text));

  Result := False;

  for i := Low(XI_True_Text) to High(XI_True_Text) do
    if UpperCase(XI_True_Text[i]) = TestStr then
    begin
      Result := True;
      Break;
    end;
end;

procedure TXmlIniKey.SetAsBool(const Value: Boolean);
begin
  if Value then
    Text := XI_True_Text[0]
  else
    Text := XI_False_Text[0];
  FDataType := xidtBool;
end;

function TXmlIniKey.GetAsDateTime: TDateTime;
begin
  Result := ExStrToDateTime(Text);
end;

procedure TXmlIniKey.SetAsDateTime(const Value: TDateTime);
begin
  Text := ExDateTimeToStr(Value);
  FDataType := xidtDateTime;
end;

function TXmlIniKey.GetAsDate: TDateTime;
begin
  Result := ExStrToDate(Text);
end;

procedure TXmlIniKey.SetAsDate(const Value: TDateTime);
begin
  Text := ExDateToStr(Value);
  FDataType := xidtDate;
end;

function TXmlIniKey.GetAsTime: TDateTime;
begin
  Result := ExStrToTime(Text);
end;

procedure TXmlIniKey.SetAsTime(const Value: TDateTime);
begin
  Text := ExTimeToStr(Value);
  FDataType := xidtTime;
end;

procedure TXmlIniKey.SetDataType(const Value: TXmlIniDataType);
var
  SS: TStringStream;
begin
  if FDataType <> Value then
  begin
    if Value = xidtBinary then
    begin
      FStream.Clear;

      try
        SS := TStringStream.Create(FText);
        try
          FStream.CopyFrom(SS, 0);
        finally
          SS.Free;
        end;
      except
        if FRaiseException then
          raise;
      end;
    end
    else if FDataType = xidtBinary then
    begin
      try
        if FStream.Size = 0 then
          FText := ''
        else
        begin
          SS := TStringStream.Create('');
          try
            SS.CopyFrom(FStream, 0);
            FText := SS.DataString;
          finally
            SS.Free;
          end;
        end;
        FStream.Clear;
      except
        if FRaiseException then
          raise;
      end;
    end;
    
    FDataType := Value;

    case FDataType of
      xidtText:
        Text := Text;
      xidtInteger:
        AsInteger := AsInteger;
      xidtFloat:
        AsFloat := AsFloat;
      xidtBool:
        AsBool := AsBool;
      xidtDateTime:
        AsDateTime := AsDateTime;
      xidtDate:
        AsDate := AsDate;
      xidtTime:
        AsTime := AsTime;
    end;
  end;
end;

{ TXmlIniEventMemoryStream }

constructor TXmlIniEventMemoryStream.Create(AOwner: TXmlIniKey);
begin
  FOwner := AOwner;
  
  inherited Create;
end;

function TXmlIniEventMemoryStream.GetSize: Int64;
begin
  if (inherited GetSize = 0) and (FOwner.FText <> '') then
    FOwner.TextToBinary;
    
  Result := inherited GetSize;
end;

procedure TXmlIniEventMemoryStream.SetSize(NewSize: Integer);
begin
  FOwner.FDataType := xidtBinary;
  FOwner.FText := '';
  
  inherited SetSize(NewSize);
end;

function TXmlIniEventMemoryStream.Read(var Buffer;
  Count: Integer): Longint;
begin
  if (inherited GetSize = 0) and (FOwner.FText <> '') then
    FOwner.TextToBinary;

  Result := inherited Read(Buffer, Count);
end;

function TXmlIniEventMemoryStream.Write(const Buffer;
  Count: Integer): Longint;
begin
  FOwner.FDataType := xidtBinary;
  FOwner.FText := '';
  
  Result := inherited Write(Buffer, Count);
end;

end.

